
#####################################################################
#
# s_util.py
#
# Project   : SAPIADAPTER
# Author(s) : Zafar Iqbal < zaf@sparc.space >
# Copyright : (C) 2022 SPARC PC < https://sparc.space/ >
#
# All rights reserved. No warranty, explicit or implicit, provided.
# SPARC PC is and remains the owner of all titles, rights
# and interests in the Software.
#
#####################################################################

import os
import sys
import time
import json
import pickle
import uuid
import hashlib
import base64
import shutil
import gzip
import inspect
import glob
import traceback
import random

import threading

from datetime import datetime , timedelta

import datetime as _dt

from subprocess import Popen, PIPE

from pathlib import Path
from traceback import format_exception

import urllib.parse

from . import get_version , s_log , s_config , s_net , s_db

#####################################################################


#class JsonEncoder( json.JSONEncoder ) :
#    def default( self , z ) :
#        if isinstance( z , datetime.date ) :
#            return ( str( z ) )
#        else :
#            return super( ).default( z )

class JsonEncoder( json.JSONEncoder ) :

    def default( self , obj ) :
        #if isinstance( obj , numpy.integer ) :
        #    return int( obj )
        #elif isinstance( obj, numpy.floating ) :
        #    return float( obj)
        #elif isinstance( obj, numpy.ndarray ) :
        #    return obj.tolist( )
        if isinstance( obj , _dt.date ) :
            return str( obj ) 
        return json.JSONEncoder.default( self , obj )

def json_encode( data ) :
    return json.dumps( data , cls = JsonEncoder )

def json_save( fn , data ) :

    #with open( fn , 'w', encoding='utf-8') as f:
    #    json.dump(data, f, ensure_ascii=False, indent=4)

    with open( fn , "w" ) as file :
        file.write( json_encode( data ) )

def json_load( fn ) :
    with open( fn ) as f :
        j = json.load( f )
    return( j )

def json_loads( str ) :
    return( json.loads( str ) )

def json_stdin( ) :

    data_in_list = sys.stdin.readlines( )
    data_in_json = "".join( data_in_list )

    data_in = json_loads( data_in_json )

    return( data_in )

def json_dumps( inputs ) :
    return( json.dumps( inputs ) )

#def json_pretty( json_in ) :
#    data = json.loads(json_in)
#    pretty_object = json.dumps(data, indent=4)
#    print(pretty_object)

def json_prettyprint( data ) :
    pretty_object = json.dumps( data , indent = 4 )
    print( pretty_object )

#####################################################################


def get_ver( ) :
    return( get_version( ) )

def num_randomlist( s = 10 ) :
    random_list = [random.random() for i in range(s)]
    return(random_list)

# Get arg from position and use default
def argv_get( p , d = None ) :
    al = len( sys.argv ) - 1
    #print( al )
    if( p > al ) : return( d )
    return( sys.argv[ p ] )

# FIXME TODO aaarrggghhh!!!
def time_getisoformat( ) :

    dt = datetime.now( datetime.now( ).astimezone( ).tzinfo ).isoformat( )

    return( dt )

#>>> now = datetime.now()
#>>> now
#datetime.datetime(2010, 11, 29, 11, 55, 27, 172480)
#>>> now.tzname()
#>>> from dateutil import tz
#>>> now = datetime.now(tz=tz.tzlocal())
#>>> now
#datetime.datetime(2010, 11, 29, 11, 56, 21, 316679, tzinfo=tzlocal())
#>>> now.tzname()

def pickle_load( datakey , fromtmp = False ) :

    fn = "local_" + datakey + ".pkl"

    if(fromtmp):
        fn="/tmp/"+fn

    if( os.path.isfile( fn ) ) :
        with open( fn , "rb" ) as f :
            return( pickle.load( f ) )
    else :
        return None
        
def pickle_save( datakey , data , mvtotmp = False ) :
    fn = "local_" + datakey + ".pkl"
    f = open( fn , "wb" )
    pickle.dump( data , f )
    f.close( )

    if( mvtotmp ) :
        os.rename( fn , "/tmp/" + fn )

def get_cwd( ) :
    directory_path = os.getcwd()
    return(directory_path)

 
def pathnotallowed( p ) :

    #s_log.write( p )

    p1 = os.path.normpath(p)
    #s_log.write( p1 )

    p2 = os.path.realpath(p1)
    #s_log.write( p2 )

    base_dir = s_config.get_key( "/config/cwd" , "/tmp")

    path_test = ( Path( base_dir ) / p2 ).resolve( ) 

    s_log.write(str(Path(base_dir).resolve()))
    s_log.write(str(path_test.parent))

    if( str( path_test.parent ).startswith( base_dir ) ) :
        return( False )

    s_log.write("pathnotallowed "+p)
    s_log.write(str(Path(base_dir).resolve()))
    s_log.write(str(path_test.parent))

    return( True )



def get_hashkeycontentdir( ) :
    # FIXME TODO should be coming from config
    fd = "/tmp/sapiadapter/sessions/"

    # FIXME TODO have this as a convenience function
    if environ_isdockerrunning( ) and dir_exists( "/data/" ) :
        fd = "/data/sessions/"

    return( fd )

def save_hashkeycontent( fk , fc ) :
    
    fd = get_hashkeycontentdir( )

    path_mkdirs( fd )

    hk = fk + s_config.get_key( "endpoint/token" )

    fp = fd + hash( hk ) + ".session"

    with open( fp , "w" ) as file :
        file.write( fc )

def load_hashkeycontent( fk ) :
    
    fd = get_hashkeycontentdir( )

    fp = fd + hash( fk ) + ".session"

    if( file_exists( fp ) ) :
        return( file_read( fp ) )
    return( False )



def uhash( prefix = "" ) :
    return( prefix + hashlib.sha256( str( uuid.uuid4( ) ).encode( "utf-8" )  ).hexdigest( ) )

# FIXME TODO bytes coming in ?!?!?
def hash( str , prefix = "" ) :
    return( prefix + hashlib.sha256( str.encode( "utf-8" ) ).hexdigest( ) )

def hash_file_md5( fp ) :
    return( hashlib.md5( open( fp , "rb" ).read( ) ).hexdigest( ) )

# FIXME TODO add encode...??? especially images...
def base64_decode( str ) :
    m = base64.b64decode( str ).decode( "utf-8" )
    return( m )


def is_ip( address ) :
    if( not address ) : return( False )
    return( not address.split( "." )[ -1 ].isalpha( ) )
    

def uptime( ) :
    return( time.time( ) - s_config.get_key( "/config/init/time" ) )


def uptime_lap( ) :

    tn = time.time( )

    tlp = s_config.get_key( "/config/init/time_lap" )  

    tl = tn - tlp 

    s_config.set_key( "/config/init/time_lap" , tn )

    return( tl )


def printexit( msg ) :
    print( "PREXIT: " + str(msg) )
    sys.exit( )


def printexception( e = False ) :
    if e :
        print( "EXCEPTION" + str( e ) )
    etype , value, tb = sys.exc_info( )
    info , error = format_exception( etype , value , tb )[ -2: ]
    print( f'Exception in:\n{info}\n{error}' )
    print( full_stack( ) )

def full_stack():
    exc = sys.exc_info()[0]
    if exc is not None:
        f = sys.exc_info()[-1].tb_frame.f_back
        stack = traceback.extract_stack(f)
    else:
        stack = traceback.extract_stack()[:-1]  # last one would be full_stack()
    trc = 'Traceback (most recent call last):\n'
    stackstr = trc + ''.join(traceback.format_list(stack))
    if exc is not None:
        # FIXME TODO WARNING this removes "chars"!
        stackstr += '  ' + traceback.format_exc().lstrip(trc)
    return stackstr


heartbeat_last = 0
def heartbeat( t = 600 ) :
    global heartbeat_last
    time.sleep( 0.01 )
    if( uptime( ) - heartbeat_last ) < t : return( False )
    s_log.write( "@" )
    heartbeat_last = uptime( )
    return( True )


def resetup_ifconfigchanged( ) :
    config = s_config.get_config( )
    if "/config/init/path_time" in config :
        cd = config[ "/config/cwd" ]
        cip = config[ "/config/init/path" ]
        cipt = config[ "/config/init/path_time" ]

        cfp = cip
        if not os.path.isabs( cip ) :
            cfp = cd + "/" + cip

        if os.path.isfile( cfp ) :

            cipt2 = os.path.getmtime( cfp )
            if( os.path.getmtime( cfp ) != cipt ) :
                s_config.setup( )
                s_db.setup( )
                s_log.write( "resetup_ifconfigchanged True" )
                #s_log.write( "CONFIG RELOADED " + cfp + str( cipt ) )


def ready( ) :

    resetup_ifconfigchanged( )

    if( not s_config.isready( ) ) :
        return( False )

    if( not s_net.check_connectivity( ) ) :
        return( False )

    return( True )



####################################################################

def gen_workingdirectory( ) :
    work_dir = wd = s_config.get_key( "sys/workspacedir" ) + "/" + uhash( "WOK" )
    return( work_dir )

####################################################################

def proc_setup( ) :

    app_dir = "/data/app"

    if( not os.path.isdir( app_dir ) ) :
        s_log.write( "app_dir " + app_dir + " does not exist!")
        return( False )


    work_dir = gen_workingdirectory( )

    # FIXME TODO should be coming from config
    shutil.copytree( app_dir , work_dir , ignore = shutil.ignore_patterns( ".git*" ) )

    os.chdir( work_dir )
    #print(os.getcwd())

    s_config.set_key( "sys/work_dir" , work_dir ) 

    # BEFORE WE ADD DEFAULT INPUT/OUTPUT FILE!!!
    proc_trackcurrentfiles( )

    ################
    #default_inputs =  s_db.meta_getdefault( )
    #s_db.files_set_defaultdata( default_inputs )
    s_db.files_defmeta_write( )
    ################

    #s_db.setup( ) 

    return( True )

def proc_trackcurrentfiles( ) :

    procfiles = { }
    for dirpath , dirs, files in os.walk( "." ) :  
        for filename in files :
            fname = os.path.join( dirpath , filename )
            #fname=os.path.abspath(fname)
            #print(fname)
            procfiles[ fname ] = True
            #if( not s_db.files_create( fname ) ) :
            #    s_log.write( "ERROR files_create " + fname )
            #with open(fname) as myfile:
            #print(myfile.read())

    s_config.set_key( "sys/procfiles" , procfiles ) 

def proc_run( ) :

    if not os.path.isfile( "_sapi.py" ) :
        s_log.write( "False isfile _sapi.py" )
        return( False )

    return( True )

def proc_storage( ) :

    procfiles = s_config.get_key( "sys/procfiles" ) 

    for dirpath , dirs, files in os.walk( "." ) :  
        for filename in files :
            fname = os.path.join( dirpath , filename )
            #fname=os.path.abspath(fname)
            #print(fname)
            if fname in procfiles : continue
            #print(fname)


            file_extension = Path( fname ).suffix

            # FIXME TODO extensions (list of) should be coming from config
            # anything starting with ".*"
            if( file_extension == ".pyc" ) : continue

            #s_log.write(fname)

            if( not s_db.files_update( fname ) ) :
                s_log.write( "ERROR files_update " + fname )
            #with open(fname) as myfile:
            #print(myfile.read())

def proc_teardown( ) :
    # FIXME TODO check a config/token flag to decide if to keep/delete work dir 
    shutil.rmtree( s_config.get_key( "sys/work_dir" )  )

    os.chdir( s_config.get_key( "/config/cwd" ) )

####################################################################

def compress( d ) :
    blob = gzip.compress( d )
    return( blob )

def decompress( d ) :
    blob = gzip.decompress( d )
    return( blob )

####################################################################

def parent_path( ) :

    # FIXME TODO this copy pasta should be cleaned up

    trc=None

    for teil in inspect.stack():
        # skip system calls
        if teil[1].startswith("<"):
            continue
        if teil[1].upper().startswith(sys.exec_prefix.upper()):
            continue
        trc = teil[1]
        
    # trc contains highest level calling script name
    # check if we have been compiled
    if getattr(sys, 'frozen', False):
        scriptdir, scriptname = os.path.split(sys.executable)
        return {"dir": scriptdir,
                "name": scriptname,
                "source": trc}


    if(trc==None):
        return(os.getcwd())

    # from here on, we are in the interpreted case
    scriptdir, trc = os.path.split(trc)
    # if trc did not contain directory information,
    # the current working directory is what we need
    if not scriptdir:
        scriptdir = os.getcwd()

    scr_dict ={"name": trc,
               "source": trc,
               "dir": scriptdir}

    #s_log.write(scr_dict)

    return scr_dict["dir"]

################

def environ_isdockerrunning( ) :
    return( "DOCKER_RUNNING" in os.environ )


def environ_getall( ) :
    return( os.environ )

def environ_get( k ) :
    if k in os.environ :
        return( os.environ[ k ] )
    return( None )

def environ_set( k , v ) :
    os.environ[ k ] = v

def container_appenvstatic_get( ) :
    ENV_STATIC = { }
    if "_CONTAINER_LOCAL_APPENVSTATIC" in os.environ :
        ENV_STATIC = dict( urllib.parse.parse_qsl( os.environ[ "_CONTAINER_LOCAL_APPENVSTATIC" ] ) )
    return( ENV_STATIC )
  

def display_time( seconds , granularity = 2 ) :

    intervals = (
        ( 'weeks' , 604800 ) , 
        ( 'days' , 86400 ) ,  
        ( 'hours' , 3600 ) , 
        ( 'minutes' , 60 ) ,
        ( 'seconds' , 1 ) ,
    )

    result = [ ]

    for name, count in intervals :
        value = seconds // count
        if value :
            seconds -= value * count
            if value == 1 :
                name = name.rstrip( 's' )
            result.append( "{} {}".format( value , name ) )
    return ', '.join( result[ :granularity ] )



def datetime_iso( ) :
    return( datetime.now( datetime.now( ).astimezone( ).tzinfo ).isoformat( ) )

def time_now( ) :
    return( time.time( ) )

def time_now_str( ) :
    return( str( round( time_now( ) ) ) )

def datetime_utcnow( ) :
    return( datetime.utcnow( ) )


def datetime_timestamp( ) :
    return( datetime.now( ).timestamp( ) )

def url_parse( u ) :
    return(urllib.parse.urlparse(u))


def sleep( t = 1 ) :
    time.sleep( t ) 

def idle( ) :
    while True:
        s_log.write( "IDLE" )
        time.sleep( 86400 )


def datetime_nowutcstr( td = 0 ) :

    d = datetime.utcnow( ) + timedelta( seconds = td)

    return( d.strftime( "%Y-%m-%dT%H:%M:%S" ) )

def datetime_now_db( td = 0 ) :

    d = datetime.now( ) + timedelta( seconds = td)

    return( d.strftime( "%Y-%m-%dT%H:%M:%S" ) )

def datetime_timedelta_db( dbdtstr , td ) :
    d = datetime.strptime( dbdtstr , "%Y-%m-%dT%H:%M:%S" ) + timedelta( seconds = td )
    return( d.strftime( "%Y-%m-%dT%H:%M:%S" ) )

def datetime_inrange_db( dbdt1 , dbdt2 , td) :
    d1 = datetime.strptime(dbdt1, '%Y-%m-%dT%H:%M:%S')  
    d2 = datetime.strptime(dbdt2, '%Y-%m-%dT%H:%M:%S') 

    dtdiff = d2 - d1
    dtdiff_seconds = int( dtdiff.total_seconds( ) )
    return( dtdiff_seconds <= td )

def datetime_diff( dt1 , dt2 ) :
        d1 = datetime.strptime( dt1 , "%Y-%m-%dT%H:%M:%S" )  
        d2 = datetime.strptime( dt2 , "%Y-%m-%dT%H:%M:%S" ) 
        dtdiff = d1 - d2 
        dt = int( dtdiff.total_seconds( ) )
        return( dt )


def datetime_diffnow( dt2 ) :
    dt1 = datetime_now_db( ) 
    return( datetime_diff( dt1 , dt2 ) )



def datetime_centerwindow( p , ec , ew ):
    a = datetime_timedelta_db( p[ ec ] , -0.5 * p[ ew ] )
    b = datetime_timedelta_db( p[ ec ] , 0.5 * p[ ew ] )
    return([a,b])


def datetime_initstartend( p , es , ee , ew ) :
    if( ( not es in p) and ( not ee in p ) ) :
        a = datetime_now_db( -1 * p[ ew ] )
        b = datetime_now_db( )
        return( [ a , b ] )
    elif( not ee in p) :
        b = datetime_timedelta_db( p[ es ] , p[ ew ] )
        return([p[es],b])
    elif( not es in p) :
        a = datetime_timedelta_db( p[ ee ] , -1 * p[ ew ] )
        return([a,p[ee]])
        
    return([p[es],p[ee]])
 


################

def path_mkdirs( dp ) :

    path = Path( dp )
    path.mkdir(parents=True,exist_ok=True)


def path_dirname( dp ) :
    return( os.path.dirname(dp) )


def dir_exists( dp ) :
    if( not os.path.isdir( dp ) ) :
        return( False )
    return( True )

def dir_chdir( dp ) :
    os.chdir( dp )

def dir_make( dp ) :
    os.makedirs(dp, exist_ok=True)
# Return file getcontents given filename and parent script __FILE__
def file_getcontent( fn , scp ) :

    fd = os.path.dirname( os.path.realpath( scp ) )

    #s_log.write(fn +","+scp)

    with open( fd + "/" + fn , mode = "rb" ) as tfile : 
        contents = tfile.read( ).decode( )

    return( contents )


def file_read(fp):
    return(open( fp , "r" ).read( ))

def file_exists( fp ) :
    return( os.path.isfile( fp ) )

def file_ctime( fp ) :
    if(not file_exists(fp)): return(False)
    return( os.path.getctime( fp ) )

def file_changed( fp , ctime ) :
    if(not file_exists(fp)): return(False)
    return( file_ctime(fp)!=ctime)

def file_rm( fp ) :
    if( not file_exists( fp ) ) : return( False )
    os.remove( fp )
    return( True )

def path_stem( p ) :
    fn = Path( p ).stem
    return( fn )


def sys_path_append_fp( fp ) :
    if( not file_exists(fp)): return(False)
    sys.path.append( os.path.dirname( fp ) )
    return(True)

def file_stat( fp ) :
    return( os.stat( fp ) )

def file_settimes( fp , t ) :
    os.utime( fp , ( t , t ) )


def file_replace( fp_source , fp_target ):
    os.replace(fp_source,fp_target)

def file_tmprandom(  ) :
    fp = "/tmp/sapitmp_" + uhash( ) + ".tmp"
    data = num_randomlist( 10000 )
    with open( fp, "w" ) as f :
        f.write( json_encode( data ) )
    return( fp )


def file_youngerthan( fp , ts ) :
    fstat = file_stat( fp )
    return( fstat.st_mtime > ts )


def dir_scan( dp ) :

    fps = [ ]
    
    if not dir_exists( dp ) :
        s_log.write("files_scan dir does not exist!")
        return( fps )

    for root , dirs , files in os.walk( dp ) :
        for f in files :
            fps.append( os.path.join( root , f ) )

    return( fps )


def matched_substring( haystack , needles ) :

    if( not needles ) : return( True )

    for needle in needles :
        if( needle in haystack ) :
            return( True )

    return( False )

################

def thread_get_id( ) :
    return( threading.get_ident( ) ) 
   
 
##################


def glob_files( dp , recursive=False ) :
    return( glob.glob( dp , recursive=recursive) )


def match_listindict( l , d ) :
    res = [ ]
    for dk in l :
        if dk in d :
            res.append( dk )

    return( list( dict.fromkeys( res ) ) )   


def math_dictkeylimit( d , k , l ) :
    if( not k in d ) :
        return( l )
    elif ( int( d[ k ] ) > l ) :
        return( l )
    return(d[k])

def dicts_merge( x , y ) :
    z = x.copy( )
    z.update( y )
    return( z )

def decode_redis(src):
    if isinstance(src, list):
        rv = list()
        for key in src:
            rv.append(decode_redis(key))
        return rv
    elif isinstance(src, dict):
        rv = dict()
        for key in src:
            rv[key.decode()] = decode_redis(src[key])
        return rv
    elif isinstance(src, bytes):
        return src.decode()
    else:
        raise Exception("type not handled: " +type(src))




#####################################################################
#
# s_cleanup.py
#
# Project   : SAPIADAPTER
# Author(s) : Zafar Iqbal < zaf@sparc.space >
# Copyright : (C) 2022 SPARC PC < https://sparc.space/ >
#
# All rights reserved. No warranty, explicit or implicit, provided.
# SPARC PC is and remains the owner of all titles, rights
# and interests in the Software.
#
#####################################################################

import atexit

#####################################################################

from . import s_log , s_db

#####################################################################

def cleanup( ) :
    # FIXME TODO only delete the threads db! otherwise another process could be still running! not all
    # FIXME TOOD count how db files are in dir and report back!
    s_db.filepath_delete( )
    #s_log.write( "cleanup" )
    pass

#####################################################################

atexit.register( cleanup )
    



#####################################################################
#
# s_confg.py
#
# Project   : SAPIADAPTER
# Author(s) : Zafar Iqbal < zaf@sparc.space >
# Copyright : (C) 2022 SPARC PC < https://sparc.space/ >
#
# All rights reserved. No warranty, explicit or implicit, provided.
# SPARC PC is and remains the owner of all titles, rights
# and interests in the Software.
#
#####################################################################

import os
import sys
import json
import time

from json.decoder import JSONDecodeError

#####################################################################

from . import get_version
from . import s_log  , s_util

#####################################################################

config = { }
config_alt = { }

ready = False

#####################################################################

def setup( config_user = None ) :

    global ready

    flag = init( config_user ) 

    if not flag :
        s_log.write( "False init!" )
        return( flag )


    preprocess( ) 

    ready = True

    s_log.write( get_config( ) )
    s_log.write( get_config_alt( ) )

    return( True )


def preprocess( ) :

    # FIXME TODO hack to flush alt config - see endpoint/url for public sapiomnibus etc...
    global config_alt
    config_alt = { }

    handle_key_fromcontainer( )
    unpack_key( )
    handle_sslverify_fornumericip( )
    handle_key_public( )
    checkup_workspacedir( )


def update_endpointkey( endpointkey ) :
    set_key( "endpoint/key" , endpointkey )
    del_key_alt( "endpoint/url" )
    unpack_key( )
    handle_sslverify_fornumericip( )
    handle_key_public( )

def update_endpointkey_fromenv( ev ) :
    update_endpointkey( s_util.environ_get( ev ) )


def get_key( config_key , default_val = None ) :

    #    config[ "/config/init/path" ] = config_path
    #    config[ "/config/init/path_time" ] = os.path.getmtime( config_path )
    # config[ "/config/cwd" ]
    if( "/config/init/path_time" in config ) :
        cd = config[ "/config/cwd" ]
        cip = config[ "/config/init/path" ]
        cipt = config[ "/config/init/path_time" ]

        cfp = cd + "/" + cip

    if( config_key in config_alt ) :
        #s_log.write_msg("GET KEY FOUND")
        if config_alt[ config_key ] != None :
            return config_alt[ config_key ]
 

    #s_log.write_msg("GET KEY "+config_key)
    if( config_key in config ) :
        #s_log.write_msg("GET KEY FOUND")
        if config[ config_key ] != None :
            return config[ config_key ]


    fsoneshot=s_util.load_hashkeycontent(config_key)
    if(fsoneshot!=False):
        return(fsoneshot)

            
    #s_log.write_msg("GET KEY NULL")
    return default_val



def checkup_workspacedir( ) :

    wd = get_key( "sys/workspacedir" )
    
    if not wd :
        set_key( "sys/workspacedir" , "/tmp/sapiadapter" )

def handle_key_fromcontainer( ) :

    if( get_key( "endpoint/key" ) ) :
        return

    if not "_CONTAINER_LOCAL_APPENDPOINTKEY" in os.environ : return

    endpoint_key = os.environ[ "_CONTAINER_LOCAL_APPENDPOINTKEY" ]
    s_log.write( "Found _CONTAINER_LOCAL_APPENDPOINTKEY" )

    if endpoint_key.endswith( "_LOCAL_APPENDPOINTKEY_" ) : return

    #set_key_alt( "endpoint/key" , endpoint_key ) 
    set_key( "endpoint/key" , endpoint_key ) 
    s_log.write( "Using _CONTAINER_LOCAL_APPENDPOINTKEY" )

def handle_key_public( ) :

    endpoint_url = get_key( "endpoint/url"  )

    #s_log.write( get_key( "endpoint/url"  ) )

    if( not ( endpoint_url == None or endpoint_url == False ) ) : return

    set_key_alt( "endpoint/url" , "https://sapiomnibus-public.sparc.space" )

def handle_sslverify_fornumericip( ) :

    # Also handles host.docker.internal

    endpoint_url = get_key( "endpoint/url" )

    if( not endpoint_url ) :
        return

    #s_log.write(endpoint_url)

    purl = s_util.url_parse( endpoint_url )
    #s_log.write(purl.scheme)
    #s_log.write(purl.hostname)
    #s_log.write(purl.port)

    if(purl.port!="443"):
        set_key_alt( "endpoint/sslverify" , "no" )
        return

    if( not purl.hostname ) :
        return

    if purl.hostname == "host.docker.internal" :
        set_key_alt( "endpoint/sslverify" , "no" )
        return

    if( s_util.is_ip( purl.hostname ) ):
        set_key_alt( "endpoint/sslverify" , "no" )
        return

    ########################################################

    host = get_key( "endpoint/host" )
    #s_log.write(host)
    if( not host ) : return

    if host.endswith(".test"):
        set_key_alt( "endpoint/sslverify" , "no" )


def unpack_key( ) :

    # TOKEN|IP:PORT|HOST
    # DEF|DEF|DEF

    endpoint_key = get_key( "endpoint/key" )

    #s_log.write(endpoint_key)

    if( not endpoint_key ) :
        endpoint_key="*|*|*"
        s_log.write( "default endpoint_key" )
        #return

    ########################################################

    key_parts = endpoint_key.split( "|" )
    key_parts_total = len( key_parts )
    #s_log.write( key_parts_total )

    ########################################################

    if( key_parts_total != 3 ) :
        s_log.write( "key_parts_total != 3 " )
        return

    ########################################################

    #if( ( key_parts[ 0 ] != "*" ) and ( len( key_parts[ 0 ] ) == 67 ) ) :
    #    #set_key_alt( "endpoint/token" , key_parts[ 0 ] )
    #   set_key( "endpoint/token" , key_parts[ 0 ] )
    set_key( "endpoint/token" , key_parts[ 0 ] )

    ########################################################

    if( ( key_parts[ 1 ] != "*" ) ) :
        url = key_parts[ 1 ]
        if( url[ 0:4 ] != "http" ) : url = "https://" + url
        purl = s_util.url_parse(url)
        if(purl.port==None):
            url=url+":443"
        #set_key_alt( "endpoint/url" , url )
        set_key( "endpoint/url" , url )

    ########################################################

    if( ( key_parts[ 2 ] != "*" ) ) :
        #set_key_alt( "endpoint/host" , key_parts[ 2 ] )
        set_key( "endpoint/host" , key_parts[ 2 ] )

    ########################################################

    # FIXME TODO do not use ALT config ?

    #s_log.write(url)
#    purl = s_util.url_parse(url)
    #s_log.write(purl.scheme)
    #s_log.write(purl.hostname)
#    s_log.write(purl.port)
#    if(purl.port==None):



    return

def isready( ) :
    return( ready )


def get_config( ) :
    # FIXME TODO merge alt config... otherwise you dont have full picture
    # FIXME TODO have log that outputs both config and alt!
    return config
def get_config_alt( ) :
    # FIXME TODO merge alt config... otherwise you dont have full picture
    # FIXME TODO have log that outputs both config and alt!
    return config_alt


def set_key( config_key , config_val ) :
    global config
    #s_log.write( "CONFIG SET_KEY " + config_key + " = " + str( config_val ) )
    config[ config_key ] = config_val
    return True


def set_key_fsoneshot( config_key , config_val ) :
    global config
    #s_log.write( "CONFIG SET_KEY " + config_key + " = " + str( config_val ) )

    if( config_key in config ) : return

    config[ config_key ] = config_val

    s_util.save_hashkeycontent( config_key , config_val )

    return True

def set_key_alt( config_key , config_val ) :
    global config_alt
    #s_log.write( "CONFIG SET_KEY " + config_key + " = " + str( config_val ) )
    config_alt[ config_key ] = config_val
    return True

def del_key_alt( kn ) :
    global config_alt
    #s_log.write_msg("GET KEY "+config_key)
    if( kn in config_alt ) :
        #s_log.write_msg("GET KEY FOUND")
        config_alt.pop( kn )
        return True
            
    #s_log.write_msg("GET KEY NULL")
    return False

def key_is_yes( kn ) :
    v = get_key( kn )
    if(v=="yes"): return(True)
    return(False)

def key_is_no( kn ) :
    v = get_key( kn )
    if( v == "no" ) : return( True )
    return( False )

def del_key( kn ) :
    #s_log.write_msg("GET KEY "+config_key)
    if( kn in config ) :
        #s_log.write_msg("GET KEY FOUND")
        config.pop( kn )
        return True
            
    #s_log.write_msg("GET KEY NULL")
    return False

################################################################

def generate_save( fp , cdict ) :
    data = {
        "sapiadapter" : cdict
    }

    s_util.json_save( fp , data )



################################################################

def init_fromfile( config_path , config_key = "sapiadapter" ) :

    global config

    #s_log.write( os.getcwd( ) + "," + config_path )

    try :

        with open( config_path ) as f :
            config_all = json.load( f )
    
    except IOError :

        s_log.write( "config init_fromfile IOError " + config_path )
        return( False )

    except JSONDecodeError as e:
        s_log.write( "JSONDecodeError config init_fromfile IOError " )
        return( False )

    ####################################################################

    if config_key in config_all :
        config.update( config_all[ config_key ] )
        #config["config/init_fromfile" ] = "yes"
        config[ "/config/init/path" ] = config_path
        config[ "/config/init/path_time" ] = os.path.getmtime( config_path )
    else :
        s_log.write( "ERROR config_key " + config_key )
        return False

    ####################################################################

    # s_log.write( "LOADED " + config_path )
    # s_log.write( config )

    return True

def load_from_file( cp ) :
    if not os.path.isabs(cp):
        s_log.write( "load_from_file file path not absolute" )
        return( False )
    if not os.path.isfile( cp ) :
        s_log.write( "load_from_file file not found" )
        return( False )

    if not init_fromfile( cp ) :
        s_log.write( "load_from_file init_fromfile false" )
        return( False )

    unpack_key( )   
    config[ "/config/init/from" ] = "manual_file"
    s_log.write( "CONFIG MANUAL" ) 
    s_log.write( config ) 
    return( True )


################################################################

# Use this when updating keys e.g. KEY otherwise session gets clobbered which may be required for thread safety
def init_update( config_user ) :

    global config

    config.update( config_user )
    config[ "/config/init/from" ] = "config_user"
    preprocess( )
    return( True ) 


def init( config_user = None ) :

    global config

    config = { }

    # Use "/..." for internal config keys...

    t1 = time.time( )
    config[ "/config/init/time" ] = t1
    config[ "/config/init/time_lap" ] = t1
    config[ "/config/version" ] = get_version( )

    config[ "/config/session" ] = s_util.uhash( )

    # FIXME TODO check usage of following key values 
    config[ "/config/cwd" ] = os.getcwd( )
    #config[ "/config/scriptdir" ] = os.path.dirname( os.path.realpath( __file__ ) )
    config[ "/config/parentpath" ] = s_util.parent_path( )

    #s_log.write( config )

    if( config_user ) :
        config.update( config_user )
        config[ "/config/init/from" ] = "config_user"
        preprocess( )
        return( True ) 

    ################################################################

    # First check env variable SAPIADAPTER
    if( "SAPIADAPTER " in os.environ ) :
        #s_log.write( "Found env var SAPIADAPTER_CONFIGURATION" )
        if( init_fromfile( os.environ[ "SAPIADAPTER_CONFIGURATION" ] ) ) :
            s_log.write( "Loaded SAPIADAPTER_CONFIGURATION " + os.environ[ "SAPIADAPTER_CONFIGURATION" ] )
            config[ "/config/init/from" ] = "environ_file"
            preprocess()
            return( True )
        else :
            return( False )

    #s_log.write( "Not found env var SAPIADAPTER_CONFIGURATION" )

    ################################################################

    config_default_filename = "local_sapiadapter_config.json"

    # Lets check current directory for config file

    if os.path.isfile( config_default_filename ) :
        if( init_fromfile( config_default_filename ) ) :
            s_log.write( "Loaded " + config_default_filename )
            config[ "/config/init/from" ] = "cwd_file"
            preprocess()
            return( True )
        return( False )

    #s_log.write( "Not found " + os.getcwd( ) + "/" + config_default_filename )

    ################################################################

    # Lets check tests directory for config file
    testpath = "tests/data/"
    if os.path.isfile( testpath + config_default_filename ) :
        if( init_fromfile( testpath + config_default_filename ) ) :
            s_log.write( "Loaded " + testpath + config_default_filename )
            config[ "/config/init/from" ] = "testsdata_file"
            preprocess()
            return( True )
        return( False )

    ################################################################

    #try:
    #    config[ "/config/maindir" ] = os.path.dirname( sys.modules[ "__main__" ].__file__ )
    #except:
    #    s_log.write("maindir main/file EXCEPTION")

    # Lets check 'main' directory for config file
    #testpath = config[ "/config/maindir" ] + "/" + config_default_filename
    #if os.path.isfile( testpath ) :
    #    if( init_fromfile( testpath ) ) :
    #        s_log.write( "Loaded " + testpath )
    #        config[ "/config/init/from" ] = "maindir_file"
    #        return( True )
    #    return( False )

    #s_log.write( "Not found " + os.getcwd( ) + "/" + testpath + config_default_filename )


    s_log.write( "No loading from config file" )

    preprocess()

    return( True )




#####################################################################
#
# sapiadapter.py
#
# Project   : SAPIADAPTER
# Author(s) : Zafar Iqbal < zaf@sparc.space >
# Copyright : (C) 2022 SPARC PC < https://sparc.space/ >
#
# All rights reserved. No warranty, explicit or implicit, provided.
# SPARC PC is and remains the owner of all titles, rights
# and interests in the Software.
#
#####################################################################

import importlib
import multiprocessing
from multiprocessing import Process
from . import s_log , s_config , s_net , s_util , s_db , s_cache , s_cleanup 

#####################################################################

#s_log.write( "    ####    ####    ####    ####\n" )

#####################################################################

s_config.setup( )
s_db.setup( )

#s_log.write( s_config.get_key( "/config/version" ) ) 

#####################################################################

def isready( ) :

    if( not s_util.ready( ) ) :
        s_log.write( "ERROR False ready" )
        return( False )

    #s_log.write( "OK" )
    return( True )

#####################################################################

def config_load_from_file( cp ) :
    return( s_config.load_from_file( cp ) )

def prexit( msg ) :
    s_util.printexit( msg )

def exitiferror( ) :
    if( geterror( ) ) :
        prexit( geterror( ) )

def defdata( raw=False ) :
    return( s_db.meta_read( "_data" , raw ) )

def defmeta( raw=False) :
    return( s_db.meta_read( "_meta" , raw ) )

def geterror( ) :
    return( s_db.meta_read( "_error" ) )

def flush( ) :
    return( s_db.flush( ) )

def reset( ) :
    return( s_db.reset( ) )

def uptime( ) :
    return( s_util.uptime( ) )
def uptime_lap( ) :
    return( s_util.uptime_lap( ) )
def log_lap( ) :
    s_log.write( s_util.uptime_lap( ) )

def heartbeat( t = 60 ) :
    return( s_util.heartbeat( t ) )

def server( runner , debug_errorprexit = False ) :

    s_log.write( "SERVER START" )

    while True :

        s_log.write( "SERVER LOOP" )

        heartbeat( )

        #####################################################

        try :
            flag_jobget = s_net.job_get( )
        except Exception as e :
            s_log.write( "Exception job_get" )
            s_util.printexception( )
            # FIXME TODO add flags to server
            if debug_errorprexit : prexit( "EXCEPTION job_get" )
            print( "Exception job_get" )
            continue
            #sys.exit( 99 )

        #####################################################
        
        if flag_jobget == False :

            #s_config.get_key( "endpoint/url" ) 
            s_log.write( "ERROR False job_get" )
            if debug_errorprexit : prexit( "ERROR False job_get" )
            print( "ERROR False job_get" )
            s_util.sleep( 10 )
            continue

        if flag_jobget == None :
            continue

        #####################################################
        #context =  s_config.get_key( "header_context" )
        #s_log.write( context )
        #qparams = context[ "qparams" ]
        #####################################################


        #if( not s_util.proc_setup( ) ) :
        #    continue

        #####################################################

        #s_util.proc_run( )

        #####################################################

        #s_util.proc_storage( )
        #s_util.proc_teardown( )

        #####################################################

        #config = { 
        #    "_env" : s_util.container_appenvstatic_get( )
        #}
        runner_output = runner( )
        s_db.meta_create( "_runner_exitflag" , runner_output )

        s_net.job_set( )

        #####################################################

        s_log.write( "OK" )

    s_log.write( "SERVER END" )



#####################################################################################################
runner_omnibus_ctime = 0
# Required to make module name scope available in function for importlib
_sapi_runner_omnibus = None
def sapirunner_omnibus_reloadifchanged( path_runner ) :
    global runner_omnibus_ctime
    try :
        if( s_util.file_changed( path_runner , runner_omnibus_ctime ) ) :
            runner_omnibus_ctime = s_util.file_ctime( path_runner )
            importlib.reload( _sapi_runner_omnibus )
            s_log.write( "path_runner RELOAD" , consoleprint = True )
            s_util.sleep( 3 )

    except Exception as e :
        s_log.write( "ERROR sapirunner_omnibus_reloadifchanged " + str( e ) , consoleprint = True )
        s_util.sleep( 3 )
        return( False )

    return( True )


def runner_omnibus( path_runner ) :

    # Required to make module name scope available in function for importlib
    global _sapi_runner_omnibus

    ###############################################################################
    # _CONTAINER_LOCAL_APPSAPIENABLED

    env_appsapienabled = s_util.environ_get("_CONTAINER_LOCAL_APPSAPIENABLED")
    #if( env_appsapienabled == None ) :
    #    s_log.write("_CONTAINER_LOCAL_APPSAPIENABLED NONE DISABLED")
    #    s_util.idle( )

    if( env_appsapienabled and env_appsapienabled.endswith( "_APPSAPIENABLED_" ) ) :
        s_log.write("_CONTAINER_LOCAL_APPSAPIENABLED ENDSWITH DISABLED")
        s_util.idle( )

    ###############################################################################

    s_log.write( "RUNNER_OMNIBUS START " + path_runner )

    if( not s_util.sys_path_append_fp( path_runner ) ) :
        s_log.write( "ERROR NOT FOUND " + path_runner , consoleprint = True )
        prexit( "ERR" )

    import _sapi_runner_omnibus

    global runner_omnibus_ctime
    runner_omnibus_ctime = s_util.file_ctime( path_runner )

    while True :

        s_log.write( "SERVER LOOP" )

        heartbeat( )

        #####################################################
        if( not sapirunner_omnibus_reloadifchanged( path_runner ) ) : continue
        #####################################################

        try :
            flag_jobget = s_net.job_get( )
        except Exception as e :
            s_log.write( "Exception job_get" , consoleprint = True )
            s_util.printexception( )
            continue

        #####################################################
        
        if flag_jobget == False :
            s_log.write( "ERROR False job_get" , consoleprint = True )
            s_util.sleep( 10 )
            continue

        if flag_jobget == None :
            continue

        #####################################################

        # FIXME TODO add try catch

        runner_exitcode = -999

        try :
           
            #runner_exitcode = _sapi_runner_omnibus._entrypoint( )

            ########################################################

            # This could change if runner does a sapi request!?!?
            jid_omnibus = s_config.get_key( "jid" )

            s_log.write( "process start" )
            prc1 = multiprocessing.Process( target = _sapi_runner_omnibus._entrypoint , daemon = True )
            prc1.start( )
            prc_starttime = s_util.uptime( )
            while prc1.is_alive( ) :
                
                ##########
                job_rping = s_net.job_rping( jid_omnibus )
                if(job_rping!=False):
                    if("state" in job_rping and job_rping["state"]=="cancel"):
                        s_log.write( "cancel job!!!!" )
                        prc1.kill( )
                        break

                    #s_log.write(job_rping)
                ##########

                s_util.sleep( 3 )
                # FIXME TODO should be configurable and use actual time not some counter
                if ( s_util.uptime( ) - prc_starttime ) > 900 :
                    s_log.write( "prc_counter!!! KILL" )
                    prc1.kill( )

            prc1.join( )
            # -9 when kill, 1 when exception/error..? 0 is OK
            runner_exitcode = prc1.exitcode
            s_log.write( "process end" )

            ########################################################

            #if( runner_exitcode != 0 ) :
            #    s_log.write( "ERROR _sapi_runner_omnibus runner_exitcode " + str( runner_exitcode ) , consoleprint = True )
            #    continue
        except Exception as e :
            #ERROR _sapi_runner_omnibus Exception module 'sapiadapter.sapiadapter' has no attribute 'write'

            s_log.write( "ERROR _sapi_runner_omnibus Exception " + str( e ) , consoleprint = True )
            s_util.printexception( e )
            s_util.sleep( 3 )
            continue

        #runner_exitcode = _sapi_runner_omnibus._entrypoint( )
        s_log.write( "runner_exitcode " + str( runner_exitcode ) , consoleprint = True )

        #####################################################

        #if runner_exitcode==0:
        s_net.job_set( )

        #####################################################

        s_log.write( "OK" )

#####################################################################################################
runner_internal_ctime = 0
# Required to make module name scope available in function for importlib
_sapi_runner_internal = None
def sapirunner_internal_reloadifchanged( path_runner ) :
    global runner_internal_ctime
    try :
        if( s_util.file_changed( path_runner , runner_internal_ctime ) ) :
            runner_internal_ctime = s_util.file_ctime( path_runner )
            importlib.reload( _sapi_runner_internal )
            s_log.write( "path_runner RELOAD" , consoleprint = True )
            s_util.sleep( 3 )

    except Exception as e :
        s_log.write( "ERROR sapirunner_internal_reloadifchanged " + str( e ) , consoleprint = True )
        s_util.sleep( 3 )
        return( False )

    return( True )


def runner_internal( path_runner ) :

    # Required to make module name scope available in function for importlib
    global _sapi_runner_internal

    s_log.write( "RUNNER_internal START " + path_runner )

    if( not s_util.sys_path_append_fp( path_runner ) ) :
        s_log.write( "ERROR NOT FOUND " + path_runner , consoleprint = True )
        prexit( "ERR" )

    import _sapi_runner_internal

    global runner_internal_ctime
    runner_internal_ctime = s_util.file_ctime( path_runner )

    while True :

        s_util.sleep( 0.1 )

        heartbeat( )

        #####################################################
        if( not sapirunner_internal_reloadifchanged( path_runner ) ) : continue
        #####################################################

        try :
            runner_exitcode = _sapi_runner_internal._entrypoint( )
            if( runner_exitcode > 0 ) :
                s_log.write( "ERROR _sapi_runner_internal runner_exitcode " + str( runner_exitcode ) , consoleprint = True )
                continue
        except Exception as e :
            s_log.write( "ERROR _sapi_runner_internal Exception " + str( e ) , consoleprint = True )
            s_util.sleep( 3 )
            continue

        #####################################################

def omnibus_ping( ) :
    run( {
        "_tests_omnibus_ping" : "y"
    } )
    return( s_db.meta_read( "_job_state" ) == "done" ) 

#####################################################################

# DATA
# TIMEOUT
def run( params = { } , wait_timeout = 60 ) :

    #if( not ready( ) ) :
    #    s_log.write( "NOT READY" )
    #    return( False )


    tid = s_util.thread_get_id()




    if( not isinstance( params , ( dict ) ) ) :
        s_log.write( "false params dict" )
        return( False )

    #####################################################################

    #s_db.flush( ) 

    #####################################################################

    #s_db.meta_create( "_params" , params )

    #####################################################################

    if not s_net.job_request( params ) :
        s_log.write( "false job_request" )
        return( False )

    #####################################################################

    jid = s_config.get_key( "jid" )

    if( jid == False ) :
        s_log.write( "false job_request jid" )
        return( False )

    #####################################################################

    #s_log.write( s_db.meta_read( "_test_run_withresponse" ))
    jr = s_net.job_responsewait( jid , wait_timeout )
    #s_log.write( s_db.meta_read( "_test_run_withresponse" ))

    if( jr == False ) :
        s_log.write( "false job_wait" )
        return( False )

    if( jr == None ) :
        s_log.write( "none job_wait" )
        return( None )

    #####################################################################

    if( s_db.stdio_has_stderr( ) ) :
        s_log.write( "ERROR stdio_has_stderr " + s_db.stdio_get_stderr( ) )
        return( None )

    #####################################################################

    #if( s_db.meta_read( "_runner_exitflag" ) == False ) :
    #    return( False )

    dd = defdata( )
    if( dd == False ) : return( True )
    return( dd )




#####################################################################
#
# s_db.py
#
# Project   : SAPIADAPTER
# Author(s) : Zafar Iqbal < zaf@sparc.space >
# Copyright : (C) 2022 SPARC PC < https://sparc.space/ >
#
# All rights reserved. No warranty, explicit or implicit, provided.
# SPARC PC is and remains the owner of all titles, rights
# and interests in the Software.
#
#####################################################################

import os
import sys
import json
import time
import platform
import hashlib
import sqlite3
from pathlib import Path

from . import s_log , s_config , s_util

#####################################################################

def setup( ) :
    s_log.write( "setup" )
    filedir_create( )

#####################################################################

def reset( ) :
    filepath_delete( )
    
def fsize( ) :
    fp = filepath_get( )

    fs = os.path.getsize( fp )

    return( fs )
    #os.path.getsize( fp )

def fsize_limited( ) :
    fs = fsize( )
    if( fs > 20971520 ) :
        s_log.write( "ERROR fsize_limited " + str( fs ) )
        return( True )
    return( False )

def db_stash( ) :

    curr_fp = filepath_get( )
    new_fp = curr_fp + "-STASH"
    os.rename( curr_fp , new_fp )
    
    data = {
        "curr_fp":curr_fp,
        "new_fp":new_fp
    }

    return( s_util.json_encode( data ) )

def db_unstash( data_json ) :
    data = s_util.json_loads( data_json )
    os.rename( data[ "new_fp" ] , data[ "curr_fp" ] )

#####################################################################

def filedir_create( ) :
    os.makedirs( filedir_get( ) , exist_ok = True )

def filedir_get( ) :
    dbworkspacedir = os.path.join( s_config.get_key( "sys/workspacedir" , "/tmp/sapiadapter" ) + "/db" , "" )
    return( dbworkspacedir )

#####################################################################

def filepath_get( ) :
    tid = str( s_util.thread_get_id( ) )
    fn = s_config.get_key( "/config/session" ) + "-" + tid + ".db"
    #s_log.write(fn)
    fp = filedir_get( ) + fn
    if not os.path.isfile( fp ) :
        dbcon = sqlite3.connect( fp )
        dbcon.commit( )
        dbcon.close( )
        init_schema( )
        init_data( )
    return( fp )

def filepath_delete( ) :
    fp = filepath_get( )
    if os.path.isfile( fp ) :
        os.remove( fp )

def filepath_delete_all( ) :
    for f in os.scandir( filedir_get( ) ) :
        os.remove( f.path )

#####################################################################

def sql_execute( sql , params_tuple = False ) :

    #s_log.write( sql )

    dbcon = sqlite3.connect( filepath_get( ) )
    cur = dbcon.cursor( )
    if params_tuple == False:
        cur.execute( sql )
    else:
        cur.execute( sql , params_tuple )
    dbcon.commit( )
    dbcon.close( )
    cur = False
    dbcon = False

def sql_execute_fetch( sql , fetch_all=False) :

    #s_log.write( sql )

    dbcon = sqlite3.connect( filepath_get( ) )
    dbcon.row_factory = lambda C , R : { c[ 0 ] : R[ i ] for i , c in enumerate( C.description ) }
    cur = dbcon.cursor( )
    cur.execute( sql )

    if fetch_all : 
        rows = cur.fetchall( )
        dbcon.close( )
        cur = False
        dbcon = False
        return( rows )
    else :
        row = cur.fetchone( )
        dbcon.close( )
        cur = False
        dbcon = False
        if( not row ) : return( False )
        return( row )

#####################################################################


def flush( ) :

    # FIXME TODO Maybe delete file?

    table_truncate( "files" ) 
    table_truncate( "env" ) 
    table_truncate( "stdio" ) 
    table_truncate( "args" ) 
    table_truncate( "stacks" ) 
    table_truncate( "meta" ) 
    table_truncate( "templates" ) 

def init_data( ) :
# FIXME TODO add host environment vars e.g. hostname/machine id etc.
    sys_create( "ver_db" , "1.0" )
    sys_create( "ver_sapi" , s_config.get_key( "/config/version" ) )
    sys_create( "serial" , "ODS042e1919c5bd8f41cd8e1a2b8e455dd8f6fb00df0c3916d1137047a7aafc4c42" )
    sys_create( "motto" , "HORAS NON NUMERO NISI SERENAS" )
    sys_create( "rev" , 1 )
    #sys_create( "sys_platform" , sys.platform )
    #sys_create( "sys_version" , sys.version )
    sys_create( "traceid" , s_log.get_traceid( ) )
    sys_create( "time_isoformat" , s_util.time_getisoformat( ) )
    sys_create( "tcreate" , time.time( ) )
    sys_create( "platform_uname" , platform.uname( ) )
    sys_create( "python_version" , platform.python_version( ) )

    sys_create( "config_dbpassthru" , s_config.get_key( "db/passthru" ) )

    #s_log.write(sys_read("time_isoformat"))

def file_contents( ) :


    dbfp = filepath_get( )
    if not os.path.isfile( dbfp ) :
        s_log.write( dbfp + " NOT exists!" )
        return( False )

    with open( dbfp , mode = "rb" ) as file : 
        db_filecontents = file.read( )


    return( db_filecontents )


def blob_set( d ) :
    blob = s_util.decompress( d )
    rewrite( blob )

def blob_get( ) :
    db_filecontents = file_contents( )
    blob = s_util.compress( db_filecontents )
    return( blob )

def rewrite( payload ) :

    if( len( payload ) == 0 ) :
        s_log.write( "ERROR payload empty??")

    # FIXME TODO technically only the filepath is required - no need to create and init db file
    dbfp = filepath_get( )
    with open( dbfp , "wb" ) as file:
        file.write( payload )


    ################################################################

    ver_sapi = sys_read( "ver_sapi" )
    ver_db = sys_read( "ver_db" )

    if( ver_sapi != s_config.get_key( "/config/version" ) ) :
        #s_log.write( "WARNING,VER_SAPI MISMATCH," + ver_sapi + "!=" + s_config.get_key( "/config/version" ) + "," + ver_db )
        s_log.write( "WARNING, VERSION MISMATCH," )

#####################################################################

def table_readall( t , extra = "" ) :
    sql = "SELECT * FROM " + t + " " + extra
    rows = sql_execute_fetch( sql , fetch_all = True )
    return( rows )

def table_truncate( t , extra = "" ) :
    sql = "DELETE from " + t + " " + extra
    sql_execute( sql )

def table_rowcount( t , extra = "") :
    sql = "SELECT COUNT(*) as count FROM " + t + " " + extra
    row = sql_execute_fetch( sql )
    return( row[ "count" ] )

def tablerow_exists( t , extra = "" ) :
    row_count = table_rowcount( t , extra ) 
    if(row_count>0): return(True)
    return(False)

#####################################################################

def init_schema( ) :

    sql_execute( "CREATE TABLE sys ( sn text UNIQUE , sv text )" )
    sql_execute( "CREATE INDEX index_sys ON sys ( sn )" )

    ################################################################

    sql_execute( "CREATE TABLE files ( fp text UNIQUE , fc blob , fs integer , fh text , fm integer )" )
    sql_execute( "CREATE INDEX index_files ON files ( fp )" )

    ################################################################

    sql_execute( "CREATE TABLE env ( eid integer PRIMARY KEY AUTOINCREMENT , en text UNIQUE , ev text )" )
    sql_execute( "CREATE INDEX index_env ON env ( eid , en )" )

    ################################################################

    sql_execute( "CREATE TABLE stdio ( sid integer PRIMARY KEY AUTOINCREMENT , sk integer , sv text )" )
    sql_execute( "CREATE INDEX index_stdio ON stdio ( sid , sk )" )

    ################################################################

    sql_execute( "CREATE TABLE args ( aid integer PRIMARY KEY AUTOINCREMENT , av text )" )
    sql_execute( "CREATE INDEX index_args ON args ( aid )" )

    ################################################################

    sql_execute( "CREATE TABLE stacks ( sid integer PRIMARY KEY AUTOINCREMENT , sk integer , sv text )" )
    sql_execute( "CREATE INDEX index_stacks ON stacks ( sid , sk )" )

    ################################################################

    sql_execute( "CREATE TABLE meta ( mn text UNIQUE , mv text )" )
    sql_execute( "CREATE INDEX index_meta ON meta ( mn )" )

    ################################################################

    sql_execute( "CREATE TABLE templates ( tid integer PRIMARY KEY AUTOINCREMENT , tp text , tk text , tv text )" )
    sql_execute( "CREATE INDEX index_templates ON templates ( tid , tp )" )

#####################################################################

def sys_create( sn , sv_in ) :

    sv = s_util.json_encode( sv_in )

    sql = "INSERT INTO sys( sn , sv ) VALUES ( ? , ? )"
    data = ( sn , sv )
    sql_execute( sql , data )
    return( True )

def sys_count( ) :
    count = table_rowcount( "sys" )
    return( count )

def sys_read( sn ) :

    sql = "SELECT * FROM sys where sn='" + sn + "' limit 1"
    row = sql_execute_fetch( sql )
    if( not row ) : return( False )
    return( json.loads( row[ "sv" ] )  )

def sys_readall( ) :
    rows = table_readall( "sys" )
    return( rows )

def sys_update( sn , sv_in ) :
    sv = s_util.json_encode( sv_in )
    sql = "UPDATE sys SET sv=? where sn=?"
    data = ( sv , sn )
    sql_execute( sql , data )

def sys_delete( sn ) :

    sql = "DELETE FROM sys where sn='" + sn + "'"
    sql_execute( sql )
    return( True )

def sys_deleteall( ) :
    table_truncate( "sys" )

def sys_rev_inc( ) :
    newrev = sys_read( "rev" ) + 1
    sys_update( "rev" , newrev )

####################################################################

def stacks_create( sv_data , sk = 0 ) :

    sv = s_util.json_encode( sv_data )

    sql = "INSERT INTO stacks ( sid , sk , sv ) VALUES ( ? , ? , ? )"
    data = ( None , sk,sv )

    sql_execute( sql , data )
    return( True )

def stacks_count( sk = 0 ) :

    extra = "where sk=" + str( sk )
    count = table_rowcount( "stacks" , extra )
    return( count )

def stacks_read( sk = 0 ) :

    sql = "SELECT * FROM stacks where sk=" + str( sk ) + " order by sid desc" 
    rows = sql_execute_fetch( sql , fetch_all=True)
    return( rows )

def stacks_readall( ) :
    rows = table_readall( "stacks" , extra = "order by sid asc" )
    return( rows )

def stacks_update( sid , sk , sv_data ) :
    sv = s_util.json_encode( sv_data )
    sql = "UPDATE stacks SET sk=? , sv=? where sid=?"
    data = ( sk , sv , sid )
    sql_execute( sql , data )

def stacks_delete( sk = 0 ) :
    sql = "DELETE FROM stacks where sk=" + str( sk ) 
    sql_execute( sql )
    return( True )

# FIXME TODO finish when specific sk is given
def stacks_deleteall( sk = False ) :
    if( sk == False ) :
        table_truncate( "stacks" )
    else:
        pass
    return(True)

def stacks_pop( sk = 0 ) :

    sql = "SELECT * FROM stacks where sk=" + str( sk ) + " order by sid desc limit 1" 
    rows = sql_execute_fetch( sql ,fetch_all=True)

    if( len( rows ) != 1 ) :
        return( False )

    row = rows[ 0 ]
   
    sid = row[ "sid" ]
    sv_dict = json.loads( row[ "sv" ] )

    sql = "DELETE FROM stacks where sid=" + str( sid ) 
    sql_execute( sql )

    return( sv_dict )    


#####################################################################

def env_create( en , ev ) :

    sql = "INSERT INTO env ( eid , en , ev ) VALUES ( null , ? , ? )"
    data = ( en , ev )
    sql_execute( sql , data )
    return(True)

def env_count( ) :
    count = table_rowcount( "env" )
    return( count )

def env_read( en ) :

    sql = "SELECT * FROM env where en='" + en + "' limit 1"
    row = sql_execute_fetch( sql )
    return( row )

def env_readall( ) :
    rows = table_readall( "env" , "order by eid asc")
    return( rows )

def env_update( en , ev ) :
    sql = "UPDATE env SET ev=? where en=?"
    data = ( ev , en )
    sql_execute( sql , data )

def env_delete( en ) :

    sql = "DELETE FROM env where en='" + en + "'"
    sql_execute( sql )
    return( True )

def env_deleteall( ) :
    table_truncate( "env" )

#####################################################################

def meta_create( mn , mv_in ) :

    # FIXME TODO something better?
    meta_delete( mn )

    mv = s_util.json_encode( mv_in )

    sql = "INSERT INTO meta ( mn , mv ) VALUES ( ? , ? )"
    data = ( mn , mv )
    sql_execute( sql , data )
    return(True)

def meta_count( ) :
    count = table_rowcount( "meta" )
    return( count )


def meta_read( mn , raw = False ) :

    sql = "SELECT * FROM meta where mn='" + mn + "' limit 1"
    row = sql_execute_fetch( sql )
    if( not row ) : return( False )
    if(raw):
        return( row[ "mv" ] )
    else:
        return( json.loads( row[ "mv" ] ) )

def meta_readall( ) :
    rows = table_readall( "meta" )
    return( rows )

def meta_update( mn , mv_in ) :
    mv = s_util.json_encode( mv_in )
    sql = "UPDATE meta SET mv=? where mn=?"
    data = ( mv , mn )
    sql_execute( sql , data )

def meta_delete( mn ) :

    sql = "DELETE FROM meta where mn='" + mn + "'"
    sql_execute( sql )
    return( True )

def meta_deleteall( ) :
    table_truncate( "meta" )

def meta_createorupdate( mn , mv ) :
    t = meta_read( mn )
    if t :
        meta_update( mn , mv )
    else :
        meta_create( mn , mv )


####################################################################

def args_create( av ) :

    sql = "INSERT INTO args ( aid , av ) VALUES ( ? , ? )"
    data = ( None , av )
    sql_execute( sql , data )
    return( True )

def args_count( ) :
    count = table_rowcount( "args" )
    return( count )

def args_read( aid ) :

    sql = "SELECT * FROM args where aid=" + aid + " limit 1"
    row = sql_execute_fetch( sql )
    return( row )

def args_readall( ) :
    rows = table_readall( "args" , "order by aid asc")
    return( rows )

def args_update( aid , av ) :
    sql = "UPDATE args SET av=? where aid=?"
    data = ( av , aid )
    sql_execute( sql , data )

def args_delete( aid ) :

    sql = "DELETE FROM args where aid=" + aid 
    sql_execute( sql )
    return( True )

def args_deleteall( ) :
    table_truncate( "args" )

####################################################################

# stdin 0 , stdout 1 , stderr 2
def stdio_create( sk , sv ) :
    sql = "INSERT INTO stdio ( sid , sk , sv ) VALUES ( null , ? , ? )"
    data = ( sk , sv )
    sql_execute( sql , data )
    return( True )

def stdio_count( sk ) :
    count = table_rowcount( "stdio" , "where sk=" + str( sk ) )
    return( count )

def stdio_read( sk ) :
    sql = "SELECT * FROM stdio where sk=" + str( sk ) + " order by sid asc" 
    rows = sql_execute_fetch( sql ,fetch_all=True)
    return( rows )

def stdio_readformatted( sk ) :
    rows = stdio_read( sk )
    buff = ""
    for r in rows:
        #buff = buff + str( r["sv"],"utf-8")+"\n"
        buff = buff + r["sv"]+"\n"
    return( buff )

def stdio_readall( ) :
    rows = table_readall( "stdio" )
    return( rows )

def stdio_update( sid , sk , sv ) :
    sql = "UPDATE stdio SET sk=? , sv=? where sid=?"
    data = ( sk , sv , sid )
    sql_execute( sql , data )

def stdio_delete( sk = 0 ) :

    sql = "DELETE FROM stdio where sk = " + str( sk ) 
    sql_execute( sql )

    return( True )    

def stdio_deleteall( ) :
    table_truncate( "stdio" )

def stdio_has_stderr( ) :

    if( stdio_count( 2 ) > 0 ) :
        return( True )

    return( False )

def stdio_get_stderr( ) :
    return( stdio_readformatted( 2 ) )

def stdio_get_stdout( ) :
    return( stdio_readformatted( 1 ) )

####################################################################

def files_create( fp , trimpath = "" ) :

    # FIXME TODO have some restriction paths allowed - this breaks when creating UID directories for work spaces
    # FIXME TODO add tests for this...
    #if( s_util.pathnotallowed( fp ) ) : return( False )

    ####################################################################

    if not os.path.isfile( fp ) :
        s_log.write( fp + " NOT found!" )
        return( False )

    fs = os.path.getsize( fp )

    with open( fp , mode = "rb" ) as file : 
        fc = file.read( )


    # FIXME TODO use better hash lib blake2b?
    fh = hashlib.md5( fc ).hexdigest( )

    ####################################################################

    fp_processed = fp

    if( trimpath != "" ) :
        fp_processed = fp[len(trimpath ):]


    statinfo = os.stat( fp )

    ####################################################################

    sql = "INSERT INTO files ( fp , fc , fs , fh , fm ) VALUES (?, ?, ?, ?,?)"
    data_tuple = ( fp_processed , fc , fs , fh , statinfo.st_mtime)

    sql_execute( sql , data_tuple )

    return( True )

def files_count( ) :
    count = table_rowcount( "files" )
    return( count )

def files_read( fp ) :

    sql =  "SELECT * FROM files where fp='" + fp + "' limit 1" 
    row = sql_execute_fetch( sql )
    return( row )

def files_readall( ) :
    rows = table_readall( "files" )
    return( rows )

# FIXME TODO def files_update !!!
# Delete current path and create a new one?!?!?!
def files_update( fp ) :
    files_delete( fp ) 
    return( files_create( fp ) )

def files_delete( fp ) :

    sql = "DELETE FROM files where fp='" + fp + "'"
    sql_execute( sql )

    return( True )  

def files_deleteall( ) :
    table_truncate( "files" )

def files_dict2json( fp , fdict ) :

    #if( s_util.pathnotallowed( fp ) ) : return( False )

    fc = s_util.json_encode( fdict )

    fs = len( fc )
    fh = hashlib.md5( fc.encode( "utf-8" ) ).hexdigest( )

    fm = s_util.time_now( )

    sql = "INSERT INTO files ( fp , fc , fs , fh ,fm) VALUES (?, ?, ?, ?,?)"
    data_tuple = ( fp , fc , fs , fh ,fm)
    sql_execute( sql , data_tuple )

    return( True )

def files_getcontent( fp ) :
    f = files_read( fp )
    if not f :
        s_log.write( "False files_getcontent" )
        return( False )
    return( f[ "fc" ] )

def files_readallmeta( ) :

    sql = "SELECT fp,fs,fh,fm FROM files" 
    rows = sql_execute_fetch( sql , fetch_all = True )
    return( rows )

def files_write( fp_source , fp_target = False ) :

    if( fp_target == False ) :
        fp_target = fp_source

    fd = files_read( fp_source )

    fp_target_tmp = fp_target + s_util.uhash( )

    with open( fp_target_tmp , "wb" ) as f :
        f.write( fd[ "fc" ] )

    s_util.file_replace( fp_target_tmp , fp_target )

    s_util.file_settimes( fp_target , fd[ "fm" ] )


def files_append( fp_source , fp_target = False ) :

    if( fp_target == False ) :
        fp_target = fp_source

    fd = files_read( fp_source )

    fp_target_tmp = fp_target + s_util.uhash( )


    ######################
    # COPY current file to tmp file
    # Append new contents
    ######################

    if s_util.file_exists( fp_target ) :
        s_log.write( "fp_target exists! " + fp_target )
        with open( fp_target , "ab" ) as f :
            f.write( fd[ "fc" ] )

    else :

        with open( fp_target_tmp , "wb" ) as f :
            f.write( fd[ "fc" ] )

        s_util.file_replace( fp_target_tmp , fp_target )

    s_util.file_settimes( fp_target , fd[ "fm" ] )


# with open( dir_cache + "local_out.png" , "wb" ) as f :
#def files_write( fp ) :
#    # FIXME TODO test this... with open( dir_cache + "local_out.png" , "wb" ) as f :
#    with open( fp , "w" , 1 ) as f :
#        f.write( files_getcontent( fp ) )

def files_restore( rpath = False , mode = "overwrite" ) :

    db_files = files_readallmeta( )
    #s_log.write( db_files )
    
    for db_file in db_files :

        fp_source = db_file[ "fp" ]

        if( rpath != False ) :
            # FIXME TODO WARNING this removes "chars"!
            fp_target = rpath.rstrip( "/" ) + "/" + fp_source
        else :
            fp_target = fp_source

        #s_log.write(fp_source+" -> "+fp_target)

        dp = s_util.path_dirname( fp_target )
        #print(dp)
        dp_path = Path( dp )
        dp_path.mkdir( parents = True , exist_ok = True )

        if mode == "overwrite" :
            files_write( fp_source , fp_target )
        
        if mode == "append" :
            files_append( fp_source , fp_target )

#        with open( fp_target , "wb" ) as f :
#            f.write( files_getcontent( fp_source ) )

#        s_util.file_settimes(fp_target,db_file["fm"])

def files_createall( pat ) :
    # Let's be clever and try to trim the pattern matching directory...

    # FIXME TODO be careful... we are only checking for wildcard "**"
    # FIXME TODO WARNING this removes "chars"!
    pat_trim = pat.rstrip( "**" )
    if( pat_trim == pat ) :
        pat_trim = ""

    files = s_util.glob_files( pat , True )
    for f in files :
        if os.path.isfile( f ) :
            files_create( f , pat_trim )



####################################################################

def templates_create( tp , tk , tv ) :
    if( s_util.pathnotallowed( tp ) ) : return( False )
    sql = "INSERT INTO templates ( tp,tk,tv) VALUES ( ? , ? , ? )"
    data = ( tp , tk , tv )
    sql_execute( sql , data )
    return(True)

def templates_count( ) :
    count = table_rowcount( "templates" )
    return( count )

def templates_read( tp ) :
    sql = "SELECT * FROM templates where tp='" + tp + "' order by tid asc" 
    rows = sql_execute_fetch( sql ,fetch_all=True)
    return( rows ) 

def templates_readall( ) :
    rows = table_readall( "templates" )
    return( rows )    

def templates_update( tp , tk , tv ) :
    sql = "UPDATE templates SET tk=? , tv=? where tp=?"
    data = ( tk , tv , tp )
    sql_execute( sql , data )

def templates_delete( tp ) :

    sql = "DELETE FROM templates where tp='" + tp + "'"
    sql_execute( sql )
    return( True )

def templates_deleteall( ) :
    table_truncate( "templates" )

